#pragma once

#ifdef _WIN32
// From winerror.h
#undef NO_ERROR
#undef FAR
#undef NEAR
#endif

#include <array>
#include <vector>
#include <string>

namespace Trueface
{
    /** Error codes returned by methods */
    enum class ErrorCode {NO_ERROR, INVALID_LICENSE, FILE_READ_FAIL, UNSUPPORTED_IMAGE_FORMAT, UNSUPPORTED_MODEL, NO_FACE_IN_FRAME, FAILED,
            COLLECTION_CREATION_ERROR, DATABASE_CONNECTION_ERROR, ENROLLMENT_ERROR, MAX_COLLECTION_SIZE_EXCEEDED, NO_RECORD_FOUND,
            NO_COLLECTION_FOUND, COLLECTION_DELETION_ERROR, EXTREME_FACE_ANGLE, FACE_TOO_CLOSE, FACE_TOO_FAR };

    /** Image color formats */
    enum class ColorCode {bgr, rgb, bgra, rgba, gray, yuv_i420, yuv_nv12};

    /** Object classes */
    enum class ObjectLabel {background, aeroplane, bicycle, bird, boat,
                    bottle, bus, car, cat, chair, cow, diningtable,
                    dog, horse, motorbike, person, pottedplant,
                    sheep, sofa, train, tvmonitor};

    /** Filters the detected faces based on score thresholds obtained from ROC curve. */
    enum class FaceDetectionFilter {
        /**
         * Filter the detected faces based on a low score threshold.
         * Limits false negatives (does not detect a face),
         * but may have more false positives (classifies a non-face as a face). */
        HIGH_RECALL,
        /**
         * Filter the detected faces based on a high score threshold.
         * Limits false positives (classifies a non-face as a face),
         * but may have more false negatives (does not detect a face). */
        HIGH_PRECISION,
        /**
         * Filter the detected faces based on a medium score threshold
         * to balance false positives and false negatives.
         * We advise using this option most of the time.
         * */
        BALANCED,
        /**
         * Do not filter the detected faces by score.
         * Will have a large number of false positives (classifies a non-face as a face). */
        UNFILTERED
    };

    /** Database Management System for storing Faceprints **/
    enum class DatabaseManagementSystem {
        /** Use sqlite backend.
         * Write Faceprints to local disk.
         * Ideal for embedded systems or use cases where only one process connects to the database.*/
        SQLITE,
        /** Use a PostgreSQL backend.
         * Ideal for distributed systems requiring synchronization. */
        POSTGRESQL,
        /** Do not write Faceprints to disk, only store in ram.
         * Warning, enrolled Faceprints will not be saved after the program terminates.
         * Switching to a new collections will also delete all enrolled templates. */
        NONE
    };

    /** Facial recognition models.
     * To compare model performances, refer to our <a href="https://docs.trueface.ai/ROC-Curves-d47d2730cf0a44afacb39aae0ed1b45a">ROC curves</a>.
     * You can also find more information on our <a href="https://reference.trueface.ai/cpp/dev/latest/py/faq.html#what-are-the-differences-between-the-face-recognition-models">FAQ page</a>.
     * The current most accurate model is TFV5. */
    enum class FacialRecognitionModel {
        /** Our most lightweight model with fastest inference time but lowest accuracy,
         * ideal for embedded systems or lightweight CPU only deployments, prototyping, and some 1 to 1 matching use cases.*/
        LITE,
        /**
         * Lightweight model which has improved accuracy over the previous LITE model, though does have slightly greater inference time.
         * Ideal for embedded systems or lightweight CPU only deployments, prototyping, and some 1 to 1 matching use cases.
         */
         LITE_V2,
        /** Full TFV4 model which has better accuracy than the LITE model, but also has greater inference time.
         * Ideal for GPU deployments and for 1 to N use cases.
         * Note, TFV4 has now been deprecated and replaced by TFV5 which has better performance.
         * Despite this, we will continue providing support for TFV4 for clients with existing collections. */
        FULL [[deprecated("Use TFV5 instead.")]],
        /** TFV5 is currently the highest accuracy model.
         * Ideal for GPU deployments and for 1 to N use cases. */
        TFV5
    };

    /** Object detection models */
    enum class ObjectDetectionModel {ACCURATE, FAST};

    /**
     * Enable GPU support (default is false).
     * Note, GPU support requires a different version of the SDK.
     * You are able to enable GPU for all modules or only specific modules.
    **/
    struct EnableGPU {
        /** Enable GPU inference for face detection */
        bool faceDetector = false;
        /** Enable GPU inference for face recognition template generation */
        bool faceRecognizer = false;

        EnableGPU() = default;
        /** Enable or disable GPU inference for all modules */
        EnableGPU(bool val) : faceDetector(val), faceRecognizer(val) {};
    };

    /**
     * Encrypt the biometric templates and identity strings when storing in the database using AES encryption.
     * Note, enabling this option does add overhead to Faceprint enrollment as well as loading a collection from a database into memory.
     * Enabling encryption does not increase the 1 to N identification time.
     */
    struct EncryptDatabase {
        /**
         * Enable database encryption.
         * Must provide encryption key if encryption is enabled.
         * If enabling encryption with PostgreSQL backend, it is strongly advised to require SSL for PostgreSQL connection.
         **/
        bool enableEncryption = false;
        /**
         * Encryption key.
         * The key is hashed to a fixed length before being used for encryption.
         * */
        std::string key;
    };

    /**
     * Initialize module in SDK constructor.
     * By default, the SDK uses lazy initialization, meaning modules are only initialized when they are first used (on first inference).
     * This is done so that modules which are not used do not load their models into memory, and hence do not utilize memory.
     * The downside to this is that the first inference will be much slower as the model file is being decrypted and loaded into memory.
     * Therefore, if you know you will use a module, choose to pre-initialize the module, which reads the model file into memory in the SDK constructor.
     */
    struct InitializeModule {
        /** Face detector */
        bool faceDetector = false;
        /** Face recognizer */
        bool faceRecognizer = false;
        /** Object detector */
        bool objectDetector = false;
        /** Bodypose estimator */
        bool bodyposeEstimator = false;
        /** Liveness */
        bool liveness = false;
        /** Active spoof */
        bool activeSpoof = false;
        /** 106 face point landmark detector */
        bool landmarkDetector = false;
    };

    /** SDK configuration options */
    struct ConfigurationOptions {
        /**
         * **Android Only**
         * Power saving mode.
         * 0 = all cores enabled (default).
         * 1 = only little clusters enabled.
         * 2 = only big clusters enabled.
        */
        int mobilePowerSave = 0;
        /**
         * **Mobile Only**
         * Set the number of threads used for inference.
         * For non-mobile platforms, should use the ``OMP_NUM_THREADS`` environment variable.
         * */
        int mobileThreads = 4;
        /**
         * **iOS Only**
         * SDK user must define how much memory is available for iOS application.
         * Should be set using the following code:
         * #include <os/proc.h>
         * options.mobileAvailableMemory = os_proc_available_memory();
         * */
        size_t mobileAvailableMemory = 0;

        /** The model to be used for facial recognition (default is TFV5) */
#if __APPLE__
    #include "TargetConditionals.h"
    #if TARGET_OS_IPHONE && TARGET_OS_SIMULATOR
        FacialRecognitionModel frModel = FacialRecognitionModel::LITE;
    #else
        FacialRecognitionModel frModel = FacialRecognitionModel::TFV5;
    #endif
#else
        FacialRecognitionModel frModel = FacialRecognitionModel::TFV5;
#endif

        /** The model to be used for object detection (default is ACCURATE model) */
        ObjectDetectionModel objModel = ObjectDetectionModel::ACCURATE;
        /**
         * The smallest face height that the face detector can detect. (default is 40 pixels, min value is 16 pixels).
         * The face detector has a detection scale range of about 5 octaves.
         * Ex. 40 pixels yields the detection scale range of ~40 pixels to 1280 (=40x2^5) pixels.
         * If set to -1, will dynamically adjusts the face detection scale range from image-height/32 to image-height
         * to ensure that large faces are detected in high resolution images.
         * Increasing the `Trueface::ConfigurationOptions.smallestFaceHeight` will result in faster face detection.
        **/
        int smallestFaceHeight = 40;
        /**
         * The face detection mode (default is VERSATILE).
         * This configuration option is now deprecated and may be removed in future releases. */
        FaceDetectionFilter fdFilter = FaceDetectionFilter::BALANCED;
        /** Database management system for storing Faceprints (default is SQLITE) **/
        DatabaseManagementSystem dbms = DatabaseManagementSystem::SQLITE;
        /** The directory path containing the model files **/
        std::string modelsPath = "./";
        /** Improves the computation speed for 1 to 1 comparisons and 1 to N searches
         * by compressing the feature vector and enabling additional optimizations. **/
        bool frVectorCompression = false;
        /**
         * Enable GPU support (default is false).
         * Note, GPU support requires a different version of the SDK.
         * You are able to enable GPU for all modules or only specific modules.
        **/
        EnableGPU enableGPU{};
        /** GPU device index */
        unsigned int deviceIndex = 0;
        /**
         * Specify the batch sizes which will be used, resulting in improved performance when switching between batch sizes.
         * If the batch sizes are specified, then using a non-specified batch size will result in an exception being thrown.
         * Note, GPU memory will be allocated for each of the specified batch sizes,
         * so specifying too many batch sizes may result in an out-of-memory crash.
         * Leave vector empty to support dynamic batch sizes. Dynamic batch sizes will result in a slowdown when switching between batch sizes.
         * Ex. If processing 36 images in the following batches: 10, 10, 10, 6 there will be a slight slowdown when switching from the batch size of 10 to 6.
         * */
        std::vector<unsigned int> batchSizes = {};
        /** Initialize specified modules in the SDK constructor (default uses lazy initialization). */
        InitializeModule initializeModule{};
        /** Encrypt the biometric templates and identity strings when storing in the database using AES encryption (default is disabled). */
        EncryptDatabase encryptDatabase{};
    };

    template <typename T>
    struct Point {
        /** Coordinate along the horizontal axis, or pixel column */
        T x;
        /** Coordinate along the vertical axis, or pixel row */
        T y;
    };

    /** Options used when generating the feature vector **/
    struct ModelOptions {
        /** Indicates if the `Trueface::ConfigurationOptions.frVectorCompression` was enabled when generating the template **/
        bool frVectorCompression = false;
    };

    /** Face feature vector and meta data  **/
    struct Faceprint {
        /** Vector of floats which describe the face **/
        std::vector<float> featureVector;
        /** SDK version used to generate feature vector **/
        std::string sdkVersion;
        /** Name of model used to generate feature vector **/
        std::string modelName;
        /** Additional options used when generating the feature vector **/
        ModelOptions modelOptions;
    };

    struct BoundingBox {
        /** The object class as a label */
        ObjectLabel label;
        /** The detection's confidence probability */
        float probability;
        /** The top left corner Point of the bounding box */
        Point<float> topLeft;
        /**  The width of the bounding box */
        float width;
        /**  The height of the bounding box */
        float height;
    };

    /** Landmark representing body joint. */
    struct Landmark {
        /** Coordinate of the joint */
        Point<float> point;
        /** Score for the joint */
        float score;
        /** Label for the joint */
        std::string label;
    };

    struct FaceBoxAndLandmarks
    {
        /**  The top left corner Point of the bounding box */
        Point<float> topLeft;
        /**  The bottom right corner Point of the bounding box */
        Point<float> bottomRight;
        /**  A vector of Point representing the facial landmarks: left eye, right eye, nose, left mouth corner, right mouth corner */
        std::vector<Point<float>> landmarks;
        /** Likelihood of this being a true positive; a value lower than 0.85 indicates a high chance of being a false positive */
        float score;
    };

    /** Array of 106 face points. */
    using Landmarks = std::array<Point<float>, 106>;

    struct Candidate {
        /** The computed similarity measure */
        float similarityMeasure;
        /** The probability the two face feature vectors are a match */
        float matchProbability;
        /** The identity of the match */
        std::string identity;
        /** The UUID of the match */
        std::string UUID;
    };

    /** Results from spoof detection */
    enum class SpoofLabel {
        /** The face image is real */
        REAL,
        /** The image may be an attempted spoof */
        FAKE
    };

    /**
     * Active spoof involves two stages: a near image and a far image.
     * The following enum is used to track which stage image is being analyzed.
     */
    enum class ActiveSpoofStage {
        /** Near image. */
        NEAR,
        /** Far image. */
        FAR
    };

    /** Results from mask detection */
    enum class MaskLabel {
        /** There is a mask on the face */
        MASK,
        /** There is no mask on the face */
        NO_MASK
    };

    /** Results from eye glasses detection */
    enum class GlassesLabel {
        /** There are glasses on the face */
        GLASSES,
        /** There are no glasses on the face */
        NO_GLASSES
    };

    /** Collection metadata */
    struct CollectionMetadata {
        /** The name of the collection */
        std::string collectionName = "";
        /** The number of unique identities in the collection */
        uint64_t numIdentities = 0;
        /** The total number of Faceprint enrolled in the collection */
        uint64_t numFaceprints = 0;
        /** The name of the face recognition model used to generate the Faceprint enrolled in the collection */
        std::string modelName = "";
        /** The size of the Faceprint feature vector in bytes */
        uint16_t featureVectorSizeBytes = 0;
        /** Additional options which were used to generate Faceprint enrolled in the collection */
        ModelOptions modelOptions {};
        /** Indicates if the collection is encrypted */
        bool encrypted = false;
    };

    /**
     * Results from blink detection.
     * You can use the members of `Trueface::BlinkState` to set your own thresholds which work best for your camera and lighting conditions.
     * */
    struct BlinkState {
        /** Predicted result for the left eye being closed.
         * Computed as `Trueface::BlinkState.leftEyeScore` < 0.3 && `Trueface::BlinkState.leftEyeAspectRatio` < 0.22. */
        bool isLeftEyeClosed = false;
        /** Predicted result for the right eye being closed.
         * Computed as `Trueface::BlinkState.rightEyeScore` < 0.3 && `Trueface::BlinkState.rightEyeAspectRatio` < 0.22. */
        bool isRightEyeClosed = false;
        /** Score indicating if the left eye is open or closed.
         * 0 indicates closed, 1 indicates open. */
        float leftEyeScore = 0.f;
        /** Score indicating if the right eye is open or closed.
         * 0 indicates closed, 1 indicates open. */
        float rightEyeScore = 0.f;
        /** The aspect ratio of the left eye, computed as height / width. */
        float leftEyeAspectRatio = 0.f;
        /** The aspect ratio of the right eye, computed as height / width. */
        float rightEyeAspectRatio = 0.f;

    };

    /**
     * The image properties.
     */
    struct ImageProperties {
        /** The image width. */
        unsigned int width = 0;
        /** The image height. */
        unsigned int height = 0;
        /** Number of channels for the image. */
        unsigned int channels = 0;
        /** Is the image stored in VRAM / GPU memory. */
        bool isGpuImage = false;
    };
}
